# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["M Sanchez del Rio"]
__license__ = "MIT"
__date__ = "27/05/2018"

"""

test write/read SRW wavefront to hdf5 files

test based on Example 4 of the SRW distribution (undulator + lens)

"""

import unittest
import numpy
from oasys_srw.srwlib import *

from wofrysrw.util.srw_hdf5 import save_wfr_2_hdf5, load_hdf5_2_wfr, load_hdf5_2_dictionary
import os


class SRWhdf5Test(unittest.TestCase):

    def create_source(self):

        #############################################################################
        # SRWLIB Example#4: Calculating synchrotron (undulator) radiation electric field (from one electron)
        # and simulating wavefront propagation through a simple optical system
        # v 0.06
        #############################################################################

        print('SRWLIB Python Example # 4:')
        print('Calculating synchrotron (undulator) radiation electric field (from one electron) and performing simulation of wavefront propagation through a simple optical system')

        #***********Undulator
        numPer = 40.5 #Number of ID Periods (without counting for terminations; semi-integer => symmetric vs long. pos.; to allow harmonics to be symmetric or anti-symmetric?)
        undPer = 0.049 #Period Length [m]
        Bx = 0.57/3. #Peak Horizontal field [T]
        By = 0.57 #Peak Vertical field [T]
        phBx = 0 #Initial Phase of the Horizontal field component
        phBy = 0 #Initial Phase of the Vertical field component
        sBx = -1 #Symmetry of the Horizontal field component vs Longitudinal position
        sBy = 1 #Symmetry of the Vertical field component vs Longitudinal position
        xcID = 0 #Transverse Coordinates of Undulator Center [m]
        ycID = 0
        zcID = 0 #Longitudinal Coordinate of Undulator Center [m]

        und = SRWLMagFldU([SRWLMagFldH(1, 'v', By, phBy, sBy, 1), SRWLMagFldH(1, 'h', Bx, phBx, sBx, 1)], undPer, numPer) #Ellipsoidal Undulator
        magFldCnt = SRWLMagFldC([und], array('d', [xcID]), array('d', [ycID]), array('d', [zcID])) #Container of all Field Elements

        #***********Electron Beam
        elecBeam = SRWLPartBeam()
        elecBeam.Iavg = 0.5 #Average Current [A]
        elecBeam.partStatMom1.x = 0. #Initial Transverse Coordinates (initial Longitudinal Coordinate will be defined later on) [m]
        elecBeam.partStatMom1.y = 0.
        elecBeam.partStatMom1.z = -0.5*undPer*(numPer + 4) #Initial Longitudinal Coordinate (set before the ID)
        elecBeam.partStatMom1.xp = 0 #Initial Relative Transverse Velocities
        elecBeam.partStatMom1.yp = 0
        elecBeam.partStatMom1.gamma = 3./0.51099890221e-03 #Relative Energy

        #***********Precision Parameters for SR calculation
        meth = 1 #SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
        relPrec = 0.01 #relative precision
        zStartInteg = 0 #longitudinal position to start integration (effective if < zEndInteg)
        zEndInteg = 0 #longitudinal position to finish integration (effective if > zStartInteg)
        npTraj = 20000 #Number of points for trajectory calculation
        useTermin = 1 #Use "terminating terms" (i.e. asymptotic expansions at zStartInteg and zEndInteg) or not (1 or 0 respectively)
        sampFactNxNyForProp = 1 #sampling factor for adjusting nx, ny (effective if > 0)
        arPrecPar = [meth, relPrec, zStartInteg, zEndInteg, npTraj, useTermin, sampFactNxNyForProp]

        #***********Initial Wavefront
        wfr = SRWLWfr()
        wfr.allocate(1, 100, 100) #Numbers of points vs Photon Energy, Horizontal and Vertical Positions (may be modified by the library!)
        wfr.mesh.zStart = 20. #Longitudinal Position [m] at which SR has to be calculated
        wfr.mesh.eStart = 1095. #1090. #Initial Photon Energy [eV]
        wfr.mesh.eFin = 1095. #1090. #Final Photon Energy [eV]
        wfr.mesh.xStart = -0.001 #Initial Horizontal Position [m]
        wfr.mesh.xFin = 0.001 #Final Horizontal Position [m]
        wfr.mesh.yStart = -0.001 #Initial Vertical Position [m]
        wfr.mesh.yFin = 0.001 #Final Vertical Position [m]
        wfr.partBeam = elecBeam


        #**********************Calculation (SRWLIB function calls) and post-processing
        print('   Performing Initial Electric Field calculation ... ', end='')
        srwl.CalcElecFieldSR(wfr, 0, magFldCnt, arPrecPar)
        print('done')

        return wfr


    def propagate_beamline(self,wfr):

        #############################################################################
        # SRWLIB Example#4: Calculating synchrotron (undulator) radiation electric field (from one electron)
        # and simulating wavefront propagation through a simple optical system
        # v 0.06
        #############################################################################


        #***********Optical Elements and Propagation Parameters
        fx = wfr.mesh.zStart/2 #Lens focal lengths
        fy = wfr.mesh.zStart/2
        optLens = SRWLOptL(fx, fy) #Lens
        optDrift = SRWLOptD(wfr.mesh.zStart) #Drift space

        propagParLens = [1, 1, 1., 0, 0, 1., 1.5, 1., 1.5, 0, 0, 0]
        propagParDrift = [1, 1, 1., 0, 0, 1., 1., 1., 1., 0, 0, 0]
        #Wavefront Propagation Parameters:
        #[0]: Auto-Resize (1) or not (0) Before propagation
        #[1]: Auto-Resize (1) or not (0) After propagation
        #[2]: Relative Precision for propagation with Auto-Resizing (1. is nominal)
        #[3]: Allow (1) or not (0) for semi-analytical treatment of the quadratic (leading) phase terms at the propagation
        #[4]: Do any Resizing on Fourier side, using FFT, (1) or not (0)
        #[5]: Horizontal Range modification factor at Resizing (1. means no modification)
        #[6]: Horizontal Resolution modification factor at Resizing
        #[7]: Vertical Range modification factor at Resizing
        #[8]: Vertical Resolution modification factor at Resizing
        #[9]: Type of wavefront Shift before Resizing (not yet implemented)
        #[10]: New Horizontal wavefront Center position after Shift (not yet implemented)
        #[11]: New Vertical wavefront Center position after Shift (not yet implemented)
        optBL = SRWLOptC([optLens, optDrift], [propagParLens, propagParDrift]) #"Beamline" - Container of Optical Elements (together with the corresponding wavefront propagation instructions)


        print('   Simulating Electric Field Wavefront Propagation ... ', end='')
        srwl.PropagElecField(wfr, optBL)
        print('done')

        return wfr

    def test_source(self):

        print("\n#\n# SRW hdf5 test write/load source\n#\n")

        wfr = self.create_source()

        save_wfr_2_hdf5(wfr,"tmp1.h5",intensity=True,phase=True,overwrite=True)

        wfr_loaded = load_hdf5_2_wfr("tmp1.h5","wfr")

        save_wfr_2_hdf5(wfr_loaded,"tmp2.h5",intensity=True,phase=True,overwrite=True)


        wf1 = load_hdf5_2_dictionary("tmp1.h5","wfr")
        wf2 = load_hdf5_2_dictionary("tmp2.h5","wfr")
        print("comparing wavefront at source")
        for key in wf1.keys():
            print("   checking field: ",key)
            numpy.testing.assert_almost_equal(wf1[key],wf2[key])

        os.remove("tmp1.h5")
        os.remove("tmp2.h5")


    def test_propagation(self):

        # calculate:
        wfr = self.create_source()

        save_wfr_2_hdf5(wfr,"tmp3.h5",intensity=True,phase=True,overwrite=True)

        wfr_end = self.propagate_beamline(wfr)

        save_wfr_2_hdf5(wfr_end,"tmp3.h5",intensity=True,phase=False,overwrite=False,subgroupname="wfr_end")

        # load:
        wfr_loaded = load_hdf5_2_wfr("tmp3.h5","wfr")

        save_wfr_2_hdf5(wfr_loaded,"tmp3bis.h5",intensity=True,phase=False,overwrite=True)

        wfr_end2 = self.propagate_beamline(wfr_loaded)

        save_wfr_2_hdf5(wfr_end2,"tmp3bis.h5",intensity=True,phase=True,overwrite=False,subgroupname="wfr_end")

        wf1_source = load_hdf5_2_dictionary("tmp3.h5","wfr")
        wf2_source = load_hdf5_2_dictionary("tmp3bis.h5","wfr")
        print("comparing wavefront at source")
        for key in wf1_source.keys():
            print("   checking field: ",key)
            numpy.testing.assert_almost_equal(wf1_source[key],wf2_source[key])

        wf1_end = load_hdf5_2_dictionary("tmp3.h5","wfr_end")
        wf2_end = load_hdf5_2_dictionary("tmp3bis.h5","wfr_end")
        print("comparing wavefront propagated")
        for key in wf1_source.keys():
            print("   checking field: ",key)
            numpy.testing.assert_almost_equal(1e-6*wf1_end[key],1e-6*wf2_end[key],1)

        os.remove("tmp3.h5")
        os.remove("tmp3bis.h5")



from oasys_srw.srwlib import srwl, SRWLWfr, SRWLRadMesh, SRWLStokes, array as srw_array

import copy
import numpy
import scipy.constants as codata

m_to_eV = codata.h*codata.c/codata.e

from wofry.propagator.wavefront import WavefrontDimension

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.decorators import WavefrontDecorator
from wofry.propagator.polarization import Polarization

from wofrysrw.srw_object import SRWObject

class WavefrontPrecisionParameters(SRWObject):
    def __init__(self,
                 sr_method = 1,  #SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
                 relative_precision = 0.01, # relative precision
                 start_integration_longitudinal_position = 0, # longitudinal position to start integration (effective if < zEndInteg)
                 end_integration_longitudinal_position = 0, # longitudinal position to finish integration (effective if > zStartInteg)
                 number_of_points_for_trajectory_calculation = 50000, #Number of points for trajectory calculation
                 use_terminating_terms = 1, # Use "terminating terms" (i.e. asymptotic expansions at zStartInteg and zEndInteg) or not (1 or 0 respectively)
                 sampling_factor_for_adjusting_nx_ny = 0.0 # sampling factor for adjusting nx, ny (effective if > 0)
                 ):
        self._sr_method = sr_method
        self._relative_precision = relative_precision
        self._start_integration_longitudinal_position = start_integration_longitudinal_position
        self._end_integration_longitudinal_position = end_integration_longitudinal_position
        self._number_of_points_for_trajectory_calculation = number_of_points_for_trajectory_calculation
        self._use_terminating_terms = use_terminating_terms
        self._sampling_factor_for_adjusting_nx_ny = sampling_factor_for_adjusting_nx_ny

    def to_SRW_array(self):
        return [int(self._sr_method),
                float(self._relative_precision),
                float(self._start_integration_longitudinal_position),
                float(self._end_integration_longitudinal_position),
                int(self._number_of_points_for_trajectory_calculation),
                int(self._use_terminating_terms),
                float(self._sampling_factor_for_adjusting_nx_ny)]

    def to_python_code(self):
        text_code  =  "[" + str(int(self._sr_method)) + ","
        text_code +=        str(float(self._relative_precision)) + ","
        text_code +=        str(float(self._start_integration_longitudinal_position)) + ","
        text_code +=        str(float(self._end_integration_longitudinal_position)) + ","
        text_code +=        str(int(self._number_of_points_for_trajectory_calculation)) + ","
        text_code +=        str(int(self._use_terminating_terms)) + ","
        text_code +=        str(float(self._sampling_factor_for_adjusting_nx_ny)) + "]"

        return text_code


class WavefrontParameters(SRWObject):
    def __init__(self, 
                 photon_energy_min = 100,
                 photon_energy_max = 10100,
                 photon_energy_points = 51,
                 h_slit_gap = 0,
                 h_slit_points = 1,
                 v_slit_gap = 0,
                 v_slit_points = 1,
                 h_position = 0.0,
                 v_position = 0.0,
                 distance = 10.0,
                 wavefront_precision_parameters=WavefrontPrecisionParameters()):
        self._photon_energy_min         = photon_energy_min
        self._photon_energy_max         = photon_energy_max
        self._photon_energy_points      = photon_energy_points
        self._h_slit_gap                = h_slit_gap
        self._h_slit_points             = h_slit_points
        self._v_slit_gap                = v_slit_gap
        self._v_slit_points             = v_slit_points
        self._h_position                = h_position
        self._v_position                = v_position
        self._distance                  = distance
        self._wavefront_precision_parameters  = wavefront_precision_parameters

    def to_SRWRadMesh(self):
        return SRWLRadMesh(_eStart=self._photon_energy_min,
                           _eFin=self._photon_energy_max,
                           _ne=int(self._photon_energy_points),
                           _xStart=self._h_position-self._h_slit_gap/2,
                           _xFin=self._h_position+self._h_slit_gap/2,
                           _nx=int(self._h_slit_points),
                           _yStart=self._v_position-self._v_slit_gap/2,
                           _yFin=self._v_position+self._v_slit_gap/2,
                           _ny=int(self._v_slit_points),
                           _zStart=self._distance)

    def to_SRWLStokes(self):
        stk = SRWLStokes()
        stk.allocate(int(self._photon_energy_points),
                     int(self._h_slit_points),
                     int(self._v_slit_points))
        stk.mesh = self.to_SRWRadMesh()

        return stk

    def to_python_code(self, data=None):
        mesh = self.to_SRWRadMesh()

        text_code  = "mesh = SRWLRadMesh(_eStart=" + str(mesh.eStart) + "," + "\n"
        text_code += "                   _eFin  =" + str(mesh.eFin  ) + "," + "\n"
        text_code += "                   _ne    =" + str(mesh.ne    ) + "," + "\n"
        text_code += "                   _xStart=" + str(mesh.xStart) + "," + "\n"
        text_code += "                   _xFin  =" + str(mesh.xFin  ) + "," + "\n"
        text_code += "                   _nx    =" + str(mesh.nx    ) + "," + "\n"
        text_code += "                   _yStart=" + str(mesh.yStart) + "," + "\n"
        text_code += "                   _yFin  =" + str(mesh.yFin  ) + "," + "\n"
        text_code += "                   _ny    =" + str(mesh.ny    ) + "," + "\n"
        text_code += "                   _zStart=" + str(mesh.zStart) + ")" + "\n"
        text_code += "\n"
        text_code += "stk = SRWLStokes()"  + "\n"
        text_code += "stk.allocate(" + str(mesh.ne) + "," + str(mesh.nx) + "," + str(mesh.ny) + ")" + "\n"
        text_code += "stk.mesh = mesh" + "\n"

        return text_code

#Meaning of Wavefront Propagation Parameters:
#[0]: Auto-Resize (1) or not (0) Before propagation
#[1]: Auto-Resize (1) or not (0) After propagation
#[2]: Relative Precision for propagation with Auto-Resizing (1. is nominal)
#[3]: Allow (1) or not (0) for semi-analytical treatment of the quadratic (leading) phase terms at the propagation
#[4]: Do any Resizing on Fourier side, using FFT, (1) or not (0)
#[5]: Horizontal Range modification factor at Resizing (1. means no modification)
#[6]: Horizontal Resolution modification factor at Resizing
#[7]: Vertical Range modification factor at Resizing
#[8]: Vertical Resolution modification factor at Resizing
#[9]: Type of wavefront Shift before Resizing (not yet implemented)
#[10]: New Horizontal wavefront Center position after Shift (not yet implemented)
#[11]: New Vertical wavefront Center position after Shift (not yet implemented)

class WavefrontPropagationParameters(SRWObject):
    def __init__(self,
                 auto_resize_before_propagation                         = 0,
                 auto_resize_after_propagation                          = 0,
                 relative_precision_for_propagation_with_autoresizing   = 1.0,
                 allow_semianalytical_treatment_of_quadratic_phase_term = 0,
                 do_any_resizing_on_fourier_side_using_fft              = 0,
                 horizontal_range_modification_factor_at_resizing       = 1.0,
                 horizontal_resolution_modification_factor_at_resizing  = 1.0,
                 vertical_range_modification_factor_at_resizing         = 1.0,
                 vertical_resolution_modification_factor_at_resizing    = 1.0,
                 type_of_wavefront_shift_before_resizing                = 0,
                 new_horizontal_wavefront_center_position_after_shift   = 0,
                 new_vertical_wavefront_center_position_after_shift     = 0):
            self._auto_resize_before_propagation                         = auto_resize_before_propagation
            self._auto_resize_after_propagation                          = auto_resize_after_propagation
            self._relative_precision_for_propagation_with_autoresizing   = relative_precision_for_propagation_with_autoresizing
            self._allow_semianalytical_treatment_of_quadratic_phase_term = allow_semianalytical_treatment_of_quadratic_phase_term
            self._do_any_resizing_on_fourier_side_using_fft              = do_any_resizing_on_fourier_side_using_fft
            self._horizontal_range_modification_factor_at_resizing       = horizontal_range_modification_factor_at_resizing
            self._horizontal_resolution_modification_factor_at_resizing  = horizontal_resolution_modification_factor_at_resizing
            self._vertical_range_modification_factor_at_resizing         = vertical_range_modification_factor_at_resizing
            self._vertical_resolution_modification_factor_at_resizing    = vertical_resolution_modification_factor_at_resizing
            self._type_of_wavefront_shift_before_resizing                = type_of_wavefront_shift_before_resizing
            self._new_horizontal_wavefront_center_position_after_shift   = new_horizontal_wavefront_center_position_after_shift
            self._new_vertical_wavefront_center_position_after_shift     = new_vertical_wavefront_center_position_after_shift

    def to_SRW_array(self):
        return [int(self._auto_resize_before_propagation),
                int(self._auto_resize_after_propagation),
                float(self._relative_precision_for_propagation_with_autoresizing),
                int(self._allow_semianalytical_treatment_of_quadratic_phase_term),
                int(self._do_any_resizing_on_fourier_side_using_fft),
                float(self._horizontal_range_modification_factor_at_resizing),
                float(self._horizontal_resolution_modification_factor_at_resizing),
                float(self._vertical_range_modification_factor_at_resizing),
                float(self._vertical_resolution_modification_factor_at_resizing),
                int(self._type_of_wavefront_shift_before_resizing),
                float(self._new_horizontal_wavefront_center_position_after_shift),
                float(self._new_vertical_wavefront_center_position_after_shift)]

    def to_python_code(self, data=None):
        text_code  =  "[" + str(int(self._auto_resize_before_propagation)) + ","
        text_code +=        str(int(self._auto_resize_after_propagation)) + ","
        text_code +=        str(float(self._relative_precision_for_propagation_with_autoresizing)) + ","
        text_code +=        str(int(self._allow_semianalytical_treatment_of_quadratic_phase_term)) + ","
        text_code +=        str(int(self._do_any_resizing_on_fourier_side_using_fft)) + ","
        text_code +=        str(float(self._horizontal_range_modification_factor_at_resizing)) + ","
        text_code +=        str(float(self._horizontal_resolution_modification_factor_at_resizing)) + ","
        text_code +=        str(float(self._vertical_range_modification_factor_at_resizing)) + ","
        text_code +=        str(float(self._vertical_resolution_modification_factor_at_resizing)) + ","
        text_code +=        str(int(self._type_of_wavefront_shift_before_resizing)) + ","
        text_code +=        str(float(self._new_horizontal_wavefront_center_position_after_shift)) + ","
        text_code +=        str(float(self._new_vertical_wavefront_center_position_after_shift)) + "]"

        return text_code

class WavefrontPropagationOptionalParameters:
    def __init__(self,
                 orientation_of_the_output_optical_axis_vector_x=0.0,
                 orientation_of_the_output_optical_axis_vector_y=0.0,
                 orientation_of_the_output_optical_axis_vector_z=0.0,
                 orientation_of_the_horizontal_base_vector_x    =0.0,
                 orientation_of_the_horizontal_base_vector_y    =0.0):
        self.orientation_of_the_output_optical_axis_vector_x=orientation_of_the_output_optical_axis_vector_x
        self.orientation_of_the_output_optical_axis_vector_y=orientation_of_the_output_optical_axis_vector_y
        self.orientation_of_the_output_optical_axis_vector_z=orientation_of_the_output_optical_axis_vector_z
        self.orientation_of_the_horizontal_base_vector_x    =orientation_of_the_horizontal_base_vector_x
        self.orientation_of_the_horizontal_base_vector_y    =orientation_of_the_horizontal_base_vector_y

    def append_to_srw_array(self, srw_array=[]):
        srw_array.append(self.orientation_of_the_output_optical_axis_vector_x)
        srw_array.append(self.orientation_of_the_output_optical_axis_vector_y)
        srw_array.append(self.orientation_of_the_output_optical_axis_vector_z)
        srw_array.append(self.orientation_of_the_horizontal_base_vector_x    )
        srw_array.append(self.orientation_of_the_horizontal_base_vector_y    )

    def append_to_python_code(self, text_code):
        text_code = text_code[:-1]

        text_code +=  "," + str(float(self.orientation_of_the_output_optical_axis_vector_x)) + ","
        text_code +=        str(float(self.orientation_of_the_output_optical_axis_vector_y)) + ","
        text_code +=        str(float(self.orientation_of_the_output_optical_axis_vector_z)) + ","
        text_code +=        str(float(self.orientation_of_the_horizontal_base_vector_x    )) + ","
        text_code +=        str(float(self.orientation_of_the_horizontal_base_vector_y    )) + "]"

        return text_code


class PolarizationComponent:
    LINEAR_HORIZONTAL  = 0
    LINEAR_VERTICAL    = 1
    LINEAR_45_DEGREES  = 2
    LINEAR_135_DEGREES = 3
    CIRCULAR_RIGHT     = 4
    CIRCULAR_LEFT      = 5
    TOTAL              = 6

    @classmethod
    def tuple(cls):
        return ["Linear Horizontal",
                "Linear Vertical",
                "Linear 45\u00b0",
                "Linear 135\u00b0",
                "Circular Right",
                "Circular Left",
                "Total"]


class CalculationType:
    SINGLE_ELECTRON_INTENSITY = 0
    MULTI_ELECTRON_INTENSITY = 1
    SINGLE_ELECTRON_FLUX = 2
    MULTI_ELECTRON_FLUX = 3
    SINGLE_ELECTRON_PHASE = 4
    SINGLE_ELECTRON_RE_E = 5
    SINGLE_ELECTRON_IM_E = 6
    SINGLE_ELECTRON_FLUENCE = 7

class TypeOfDependence:
    VS_E = 0
    VS_X = 1
    VS_Y = 2
    VS_XY = 3
    VS_EX = 4
    VS_EY = 5
    VS_EXY = 6
'''
:param polarization_component_to_be_extracted:
               =0 -Linear Horizontal;
               =1 -Linear Vertical;
               =2 -Linear 45 degrees;
               =3 -Linear 135 degrees;
               =4 -Circular Right;
               =5 -Circular Left;
               =6 -Total
:param calculation_type:
               =0 -"Single-Electron" Intensity;
               =1 -"Multi-Electron" Intensity;
               =2 -"Single-Electron" Flux;
               =3 -"Multi-Electron" Flux;
               =4 -"Single-Electron" Radiation Phase;
               =5 -Re(E): Real part of Single-Electron Electric Field;
               =6 -Im(E): Imaginary part of Single-Electron Electric Field;
               =7 -"Single-Electron" Intensity, integrated over Time or Photon Energy (i.e. Fluence)
:param type_of_dependence:
               =0 -vs e (photon energy or time);
               =1 -vs x (horizontal position or angle);
               =2 -vs y (vertical position or angle);
               =3 -vs x&y (horizontal and vertical positions or angles);
               =4 -vs e&x (photon energy or time and horizontal position or angle);
               =5 -vs e&y (photon energy or time and vertical position or angle);
               =6 -vs e&x&y (photon energy or time, horizontal and vertical positions or angles);
:param fixed_input_photon_energy_or_time: input photon energy [eV] or time [s] to keep fixed (to be taken into account for dependences vs x, y, x&y)
:param fixed_horizontal_position: input horizontal position [m] to keep fixed (to be taken into account for dependences vs e, y, e&y)
:param fixed_vertical_position: input vertical position [m] to keep fixed (to be taken into account for dependences vs e, x, e&x)
'''
class FluxCalculationParameters(object):
    def __init__(self,
                 polarization_component_to_be_extracted=PolarizationComponent.TOTAL,
                 calculation_type=CalculationType.SINGLE_ELECTRON_INTENSITY,
                 type_of_dependence=TypeOfDependence.VS_E,
                 fixed_input_photon_energy_or_time = 0.0,
                 fixed_horizontal_position = 0.0,
                 fixed_vertical_position = 0.0):
    
        self._polarization_component_to_be_extracted = polarization_component_to_be_extracted
        self._calculation_type                       = calculation_type                      
        self._type_of_dependence                     = type_of_dependence                    
        self._fixed_input_photon_energy_or_time      = fixed_input_photon_energy_or_time     
        self._fixed_horizontal_position              = fixed_horizontal_position             
        self._fixed_vertical_position                = fixed_vertical_position     

class SRWWavefront(SRWLWfr, WavefrontDecorator):
    class ScanningData(object):

        def __init__(self,
                     scanned_variable_name,
                     scanned_variable_value,
                     scanned_variable_display_name,
                     scanned_variable_um):
            self.__scanned_variable_name = scanned_variable_name
            self.__scanned_variable_value = scanned_variable_value
            self.__scanned_variable_display_name = scanned_variable_display_name
            self.__scanned_variable_um = scanned_variable_um

        def get_scanned_variable_name(self):
            return self.__scanned_variable_name

        def get_scanned_variable_value(self):
            return self.__scanned_variable_value

        def get_scanned_variable_display_name(self):
            return self.__scanned_variable_display_name

        def get_scanned_variable_um(self):
            return self.__scanned_variable_um

    def __init__(self,
                 _arEx=None,
                 _arEy=None,
                 _typeE='f',
                 _eStart=0,
                 _eFin=0,
                 _ne=0,
                 _xStart=0,
                 _xFin=0,
                 _nx=0,
                 _yStart=0,
                 _yFin=0,
                 _ny=0,
                 _zStart=0,
                 _partBeam=None):

        SRWLWfr.__init__(self,
                         _arEx=_arEx,
                         _arEy=_arEy,
                         _typeE=_typeE,
                         _eStart=_eStart,
                         _eFin=_eFin,
                         _ne=_ne,
                         _xStart=_xStart,
                         _xFin=_xFin,
                         _nx=_nx,
                         _yStart=_yStart,
                         _yFin=_yFin,
                         _ny=_ny,
                         _zStart=_zStart,
                         _partBeam=_partBeam)

        self.scanned_variable_data = None

    def get_wavelength(self):
        if (self.mesh.eFin + self.mesh.eStart) == 0:
            return 0.0
        else:
            return m_to_eV/((self.mesh.eFin + self.mesh.eStart)*0.5)

    def get_photon_energy(self):
        if (self.mesh.eFin + self.mesh.eStart) == 0:
            return 0.0
        else:
            return (self.mesh.eFin + self.mesh.eStart)*0.5

    def get_dimension(self):
        return WavefrontDimension.TWO

    def toGenericWavefront(self):
        sigma_polarization_field, pi_polarization_field = SRWEFieldAsNumpy(srwwf=self)

        if self.mesh.ne > 1:
            wavefront = numpy.full(self.mesh.ne, None)

            step = (self.mesh.eFin - self.mesh.eStart)/self.mesh.ne

            is_time = self.mesh.eStart < 1e-7

            for index in range(self.mesh.ne):
                energy = self.mesh.eStart + step*index

                if is_time:
                    wavelenght = energy
                else:
                    wavelenght = m_to_eV/energy

                wavefront[index] = GenericWavefront2D.initialize_wavefront_from_range(self.mesh.xStart,
                                                                                      self.mesh.xFin,
                                                                                      self.mesh.yStart,
                                                                                      self.mesh.yFin,
                                                                                      number_of_points=(self.mesh.nx, self.mesh.ny),
                                                                                      wavelength=wavelenght,
                                                                                      polarization=Polarization.TOTAL)

                wavefront[index].set_complex_amplitude(sigma_polarization_field[index, :, :],
                                                       pi_polarization_field[index, :, :])
        else:
            wavefront = GenericWavefront2D.initialize_wavefront_from_range(self.mesh.xStart,
                                                                           self.mesh.xFin,
                                                                           self.mesh.yStart,
                                                                           self.mesh.yFin,
                                                                           number_of_points=(self.mesh.nx, self.mesh.ny),
                                                                           wavelength=self.get_wavelength(),
                                                                           polarization=Polarization.TOTAL)

            wavefront.set_complex_amplitude(sigma_polarization_field[0, :, :],
                                            pi_polarization_field[0, :, :])

        return wavefront

    @classmethod
    def fromGenericWavefront(cls, wavefront):

        if wavefront.is_polarized():
            return SRWWavefrontFromElectricField(horizontal_start  = wavefront.get_coordinate_x()[0],
                                                 horizontal_end    = wavefront.get_coordinate_x()[-1],
                                                 horizontal_efield = wavefront.get_complex_amplitude(polarization=Polarization.SIGMA),
                                                 vertical_start    = wavefront.get_coordinate_y()[0],
                                                 vertical_end      = wavefront.get_coordinate_y()[-1],
                                                 vertical_efield   = wavefront.get_complex_amplitude(polarization=Polarization.PI),
                                                 energy_min        = wavefront.get_photon_energy(),
                                                 energy_max        = wavefront.get_photon_energy(),
                                                 energy_points     = 1,
                                                 z                 = 0.0,
                                                 Rx                = 1e5,
                                                 dRx               = 1.0,
                                                 Ry                = 1e5,
                                                 dRy               = 1.0)
        else:
            return SRWWavefrontFromElectricField(horizontal_start  = wavefront.get_coordinate_x()[0],
                                                 horizontal_end    = wavefront.get_coordinate_x()[-1],
                                                 horizontal_efield = wavefront.get_complex_amplitude(polarization=Polarization.SIGMA),
                                                 vertical_start    = wavefront.get_coordinate_y()[0],
                                                 vertical_end      = wavefront.get_coordinate_y()[-1],
                                                 vertical_efield   = numpy.zeros_like(wavefront.get_complex_amplitude()),
                                                 energy_min        = wavefront.get_photon_energy(),
                                                 energy_max        = wavefront.get_photon_energy(),
                                                 energy_points     = 1,
                                                 z                 = 0.0,
                                                 Rx                = 1e5,
                                                 dRx               = 1.0,
                                                 Ry                = 1e5,
                                                 dRy               = 1.0)

    @classmethod
    def decorateSRWWF(self, srwwf):
        dim_x = srwwf.mesh.nx
        dim_y = srwwf.mesh.ny
        number_energies = srwwf.mesh.ne

        x_polarization = SRWArrayToNumpy(srwwf.arEx, dim_x, dim_y, number_energies)
        y_polarization = SRWArrayToNumpy(srwwf.arEy, dim_x, dim_y, number_energies)

        wavefront = SRWWavefrontFromElectricField(horizontal_start=srwwf.mesh.xStart,
                                                  horizontal_end=srwwf.mesh.xFin,
                                                  horizontal_efield=x_polarization,
                                                  vertical_start=srwwf.mesh.yStart,
                                                  vertical_end=srwwf.mesh.yFin,
                                                  vertical_efield=y_polarization,
                                                  energy_min=srwwf.mesh.eStart,
                                                  energy_max=srwwf.mesh.eFin,
                                                  energy_points=srwwf.mesh.ne,
                                                  z=srwwf.mesh.zStart,
                                                  Rx=srwwf.Rx,
                                                  dRx=srwwf.dRx,
                                                  Ry=srwwf.Ry,
                                                  dRy=srwwf.dRy)


        wavefront.numTypeElFld=srwwf.numTypeElFld
        wavefront.partBeam=srwwf.partBeam

        # debug srio@esrf.eu: added these fields as they are not correctly pased
        wavefront.mesh.nx = dim_x
        wavefront.mesh.ny = dim_y

        return wavefront


    def duplicate(self):
        wavefront = SRWWavefront(_arEx=copy.deepcopy(self.arEx),
                                 _arEy=copy.deepcopy(self.arEy),
                                 _typeE=self.numTypeElFld,
                                 _eStart=self.mesh.eStart,
                                 _eFin=self.mesh.eFin,
                                 _ne=self.mesh.ne,
                                 _xStart=self.mesh.xStart,
                                 _xFin=self.mesh.xFin,
                                 _nx=self.mesh.nx,
                                 _yStart=self.mesh.yStart,
                                 _yFin=self.mesh.yFin,
                                 _ny=self.mesh.ny,
                                 _zStart=self.mesh.zStart,
                                 _partBeam=self.partBeam)

        wavefront.mesh = copy.deepcopy(self.mesh)
        wavefront.Rx  = self.Rx
        wavefront.Ry  = self.Ry
        wavefront.dRx  = self.dRx
        wavefront.dRy  = self.dRy
        wavefront.xc  = self.xc
        wavefront.yc = self.yc
        wavefront.avgPhotEn  = self.avgPhotEn
        wavefront.presCA = self.presCA
        wavefront.presFT  = self.presFT
        wavefront.unitElFld  = self.unitElFld
        wavefront.arElecPropMatr  = copy.deepcopy(self.arElecPropMatr)
        wavefront.arMomX  = copy.deepcopy(self.arMomX)
        wavefront.arMomY  = copy.deepcopy(self.arMomY)
        wavefront.arWfrAuxData  = copy.deepcopy(self.arWfrAuxData)

        wavefront.scanned_variable_data = self.scanned_variable_data

        return wavefront

    def setScanningData(self, scanned_variable_data=ScanningData(None, None, None, None)):
        self.scanned_variable_data=scanned_variable_data

    def get_intensity(self, multi_electron=True, polarization_component_to_be_extracted=PolarizationComponent.TOTAL, type_of_dependence=TypeOfDependence.VS_XY):
        if type_of_dependence not in (TypeOfDependence.VS_X, TypeOfDependence.VS_Y, TypeOfDependence.VS_XY):
            raise ValueError("Wrong Type of Dependence: only vs. X, vs. Y, vs. XY are supported")

        if multi_electron:
            flux_calculation_parameters=FluxCalculationParameters(calculation_type   = CalculationType.MULTI_ELECTRON_INTENSITY,
                                                                  polarization_component_to_be_extracted=polarization_component_to_be_extracted,
                                                                  type_of_dependence = type_of_dependence)
        else:
            flux_calculation_parameters=FluxCalculationParameters(calculation_type   = CalculationType.SINGLE_ELECTRON_INTENSITY,
                                                                  polarization_component_to_be_extracted=polarization_component_to_be_extracted,
                                                                  type_of_dependence = type_of_dependence)
        if type_of_dependence == TypeOfDependence.VS_XY:
            return self.get_2D_intensity_distribution(type='f', flux_calculation_parameters=flux_calculation_parameters)
        elif (type_of_dependence == TypeOfDependence.VS_X or type_of_dependence == TypeOfDependence.VS_Y):
            return self.get_1D_intensity_distribution(type='f', flux_calculation_parameters=flux_calculation_parameters)

    def get_phase(self, polarization_component_to_be_extracted=PolarizationComponent.TOTAL):
        flux_calculation_parameters=FluxCalculationParameters(calculation_type   = CalculationType.SINGLE_ELECTRON_PHASE,
                                                              polarization_component_to_be_extracted=polarization_component_to_be_extracted,
                                                              type_of_dependence = TypeOfDependence.VS_XY)

        return self.get_2D_intensity_distribution(type='d', flux_calculation_parameters=flux_calculation_parameters)

    def get_flux(self, multi_electron=True, polarization_component_to_be_extracted=PolarizationComponent.TOTAL):
        if multi_electron:
            flux_calculation_parameters=FluxCalculationParameters(calculation_type   = CalculationType.MULTI_ELECTRON_INTENSITY,
                                                                  type_of_dependence = TypeOfDependence.VS_E,
                                                                  polarization_component_to_be_extracted=polarization_component_to_be_extracted,
                                                                  fixed_input_photon_energy_or_time=self.mesh.eStart,
                                                                  fixed_horizontal_position=self.mesh.xStart,
                                                                  fixed_vertical_position=self.mesh.yStart)
        else:
            flux_calculation_parameters=FluxCalculationParameters(calculation_type   = CalculationType.SINGLE_ELECTRON_INTENSITY,
                                                                  type_of_dependence = TypeOfDependence.VS_E,
                                                                  polarization_component_to_be_extracted=polarization_component_to_be_extracted,
                                                                  fixed_input_photon_energy_or_time=self.mesh.eStart,
                                                                  fixed_horizontal_position=self.mesh.xStart,
                                                                  fixed_vertical_position=self.mesh.yStart)

        output_array = srw_array('f', [0]*self.mesh.ne)

        SRWWavefront.get_intensity_from_electric_field(output_array, self, flux_calculation_parameters)
        
        data = numpy.ndarray(buffer=output_array, shape=self.mesh.ne, dtype=output_array.typecode)

        energy_array=numpy.linspace(self.mesh.eStart,
                                    self.mesh.eFin,
                                    self.mesh.ne)
        spectral_flux_array = numpy.zeros(energy_array.size)

        for ie in range(energy_array.size):
            spectral_flux_array[ie] = data[ie]

        return (energy_array, spectral_flux_array)

    def get_2D_intensity_distribution(self, type='f', flux_calculation_parameters=FluxCalculationParameters()):
        mesh = copy.deepcopy(self.mesh)

        h_array = numpy.linspace(mesh.xStart, mesh.xFin, mesh.nx)
        v_array = numpy.linspace(mesh.yStart, mesh.yFin, mesh.ny)
        e_array = numpy.linspace(mesh.eStart, mesh.eFin, mesh.ne)

        intensity_array = numpy.zeros((e_array.size, h_array.size, v_array.size))

        for ie in range(e_array.size):
            output_array = srw_array(type, [0] * mesh.nx * mesh.ny)  # "flat" array to take 2D intensity data

            flux_calculation_parameters._fixed_input_photon_energy_or_time = e_array[ie]

            SRWWavefront.get_intensity_from_electric_field(output_array, self, flux_calculation_parameters)

            # FROM UTI_PLOT in SRW
            tot_len = int(mesh.ny * mesh.nx)
            len_output_array = len(output_array)

            if len_output_array > tot_len:
                output_array = numpy.array(output_array[0:tot_len])
            elif len_output_array < tot_len:
                aux_array = srw_array('d', [0] * len_output_array)
                for i in range(len_output_array): aux_array[i] = output_array[i]
                output_array = numpy.array(srw_array(aux_array))
            else:
                output_array = numpy.array(output_array)

            output_array = output_array.reshape(mesh.ny, mesh.nx)

            for ix in range(mesh.nx):
                for iy in range(mesh.ny):
                    intensity_array[ie, ix, iy] = output_array[iy, ix]

        return (e_array, h_array, v_array, intensity_array)

    def get_1D_intensity_distribution(self, type='f', flux_calculation_parameters=FluxCalculationParameters()):
        mesh = copy.deepcopy(self.mesh)

        if flux_calculation_parameters._type_of_dependence == TypeOfDependence.VS_X:
            pos_array = numpy.linspace(mesh.xStart, mesh.xFin, mesh.nx)
        else:
            pos_array = numpy.linspace(mesh.yStart, mesh.yFin, mesh.ny)

        e_array = numpy.linspace(mesh.eStart, mesh.eFin, mesh.ne)

        intensity_array = numpy.zeros((e_array.size, pos_array.size))

        for ie in range(e_array.size):
            output_array = srw_array(type, [0] * len(pos_array))  # "flat" array to take 2D intensity data

            flux_calculation_parameters._fixed_input_photon_energy_or_time = e_array[ie]

            SRWWavefront.get_intensity_from_electric_field(output_array, self, flux_calculation_parameters)

            # FROM UTI_PLOT in SRW
            tot_len = len(pos_array)
            len_output_array = len(output_array)

            if len_output_array > tot_len:
                output_array = numpy.array(output_array[0:tot_len])
            elif len_output_array < tot_len:
                aux_array = srw_array('d', [0] * len_output_array)
                for i in range(len_output_array): aux_array[i] = output_array[i]
                output_array = numpy.array(srw_array(aux_array))
            else:
                output_array = numpy.array(output_array)

            for i in range(len(pos_array)):
                intensity_array[ie, i] = output_array[i]

        return (e_array, pos_array, intensity_array)

    @classmethod
    def get_intensity_from_electric_field(cls,
                                          output_array,
                                          srw_wavefront,
                                          flux_calculation_parameters = FluxCalculationParameters()):

        srwl.CalcIntFromElecField(output_array,
                                  srw_wavefront,
                                  flux_calculation_parameters._polarization_component_to_be_extracted,
                                  flux_calculation_parameters._calculation_type,
                                  flux_calculation_parameters._type_of_dependence,
                                  flux_calculation_parameters._fixed_input_photon_energy_or_time,
                                  flux_calculation_parameters._fixed_horizontal_position,
                                  flux_calculation_parameters._fixed_vertical_position)

        return output_array

# ------------------------------------------------------------------
# ------------------------------------------------------------------
# ------------------------------------------------------------------
# ACCESSORIES
# ------------------------------------------------------------------
# ------------------------------------------------------------------

def SRWEFieldAsNumpy(srwwf):
    """
    Extracts electrical field from a SRWWavefront
    :param srw_wavefront: SRWWavefront to extract electrical field from.
    :return: 4D numpy array: [energy, horizontal, vertical, polarisation={0:horizontal, 1: vertical}]
    """

    dim_x = srwwf.mesh.nx
    dim_y = srwwf.mesh.ny
    number_energies = srwwf.mesh.ne

    x_polarization = SRWArrayToNumpy(srwwf.arEx, dim_x, dim_y, number_energies)
    y_polarization = SRWArrayToNumpy(srwwf.arEy, dim_x, dim_y, number_energies)

    return [x_polarization, y_polarization]

def SRWWavefrontFromElectricField(horizontal_start,
                                  horizontal_end,
                                  horizontal_efield,
                                  vertical_start,
                                  vertical_end,
                                  vertical_efield,
                                  energy_min,
                                  energy_max,
                                  energy_points,
                                  z,
                                  Rx,
                                  dRx,
                                  Ry,
                                  dRy):
    """
    Creates a SRWWavefront from pi and sigma components of the electrical field.
    :param horizontal_start: Horizontal start position of the grid in m
    :param horizontal_end: Horizontal end position of the grid in m
    :param horizontal_efield: The pi component of the complex electrical field
    :param vertical_start: Vertical start position of the grid in m
    :param vertical_end: Vertical end position of the grid in m
    :param vertical_efield: The sigma component of the complex electrical field
    :param energy: Energy in eV
    :param z: z position of the wavefront in m
    :param Rx: Instantaneous horizontal wavefront radius
    :param dRx: Error in instantaneous horizontal wavefront radius
    :param Ry: Instantaneous vertical wavefront radius
    :param dRy: Error in instantaneous vertical wavefront radius
    :return: A wavefront usable with SRW.
    """

    horizontal_size = horizontal_efield.shape[0]
    vertical_size = horizontal_efield.shape[1]

    if horizontal_size % 2 == 1 or \
       vertical_size % 2 == 1:
        # raise Exception("Both horizontal and vertical grid must have even number of points")
        print("NumpyToSRW: WARNING: Both horizontal and vertical grid must have even number of points")

    horizontal_field = numpyArrayToSRWArray(horizontal_efield)
    vertical_field = numpyArrayToSRWArray(vertical_efield)

    srwwf = SRWWavefront(_arEx=horizontal_field,
                         _arEy=vertical_field,
                         _typeE='f',
                         _eStart=energy_min,
                         _eFin=energy_max,
                         _ne=energy_points,
                         _xStart=horizontal_start,
                         _xFin=horizontal_end,
                         _nx=horizontal_size,
                         _yStart=vertical_start,
                         _yFin=vertical_end,
                         _ny=vertical_size,
                         _zStart=z)

    srwwf.Rx = Rx
    srwwf.Ry = Ry
    srwwf.dRx = dRx
    srwwf.dRy = dRy

    return srwwf

def numpyArrayToSRWArray(numpy_array):
    """
    Converts a numpy.array to an array usable by SRW.
    :param numpy_array: a 2D numpy array
    :return: a 2D complex SRW array
    """
    elements_size = numpy_array.size

    r_horizontal_field = numpy_array[:, :].real.transpose().flatten().astype(numpy.float)
    i_horizontal_field = numpy_array[:, :].imag.transpose().flatten().astype(numpy.float)

    tmp = numpy.zeros(elements_size * 2, dtype=numpy.float32)
    for i in range(elements_size):
        tmp[2*i] = r_horizontal_field[i]
        tmp[2*i+1] = i_horizontal_field[i]

    return srw_array('f', tmp)

def SRWArrayToNumpy(srw_array, dim_x, dim_y, number_energies):
    """
    Converts a SRW array to a numpy.array.
    :param srw_array: SRW array
    :param dim_x: size of horizontal dimension
    :param dim_y: size of vertical dimension
    :param number_energies: Size of energy dimension
    :return: 4D numpy array: [energy, horizontal, vertical, polarisation={0:horizontal, 1: vertical}]
    """
    re = numpy.array(srw_array[::2], dtype=numpy.float)
    im = numpy.array(srw_array[1::2], dtype=numpy.float)

    e = re + 1j * im
    e = e.reshape((dim_y,
                   dim_x,
                   number_energies,
                   1)
                  )

    e = e.swapaxes(0, 2)

    return e.copy()




from oasys_srw.srwlib import array, SRWLMagFld3D, SRWLMagFldH, SRWLMagFldU

from syned.storage_ring.magnetic_structures.wiggler import Wiggler
from wofrysrw.storage_ring.srw_magnetic_structure import SRWMagneticStructure

import numpy

class SRWWiggler(Wiggler, SRWMagneticStructure):

    def __init__(self,
                 K_vertical=0.0,
                 K_horizontal=0.0,
                 period_length = 0.0,
                 number_of_periods = 1):
        Wiggler.__init__(self,
                         K_vertical=K_vertical,
                         K_horizontal=K_horizontal,
                         period_length=period_length,
                         number_of_periods=number_of_periods)

    def get_SRWMagneticStructure(self):
        if self._number_of_periods <=3:
            n_points = 100

            arBx = array('d', [0]*n_points) if self._K_horizontal > 0.0 else None
            arBy = array('d', [0]*n_points) if self._K_vertical > 0.0 else None

            longitudinal_mesh = numpy.linspace(-self._period_length/2., self._period_length/2., n_points)
            phases = 2.0*numpy.pi*(longitudinal_mesh/self._period_length)

            for i in range (0, n_points):
                if self._K_horizontal > 0.0: arBx[i] = -self.magnetic_field_horizontal() * numpy.cos(phases[i])
                if self._K_vertical > 0.0:   arBy[i] = self.magnetic_field_vertical() * numpy.cos(phases[i])

            return SRWLMagFld3D(_arBx=arBx,
                                _arBy=arBy,
                                _nx=1,
                                _ny=1,
                                _nz=n_points,
                                _rz=self._period_length,
                                _nRep=int(self._number_of_periods),
                                _interp=1)
        else:
            magnetic_fields = []

            if self._K_vertical > 0.0:
                magnetic_fields.append(SRWLMagFldH(1, 'v', self.magnetic_field_vertical(), 0, 1, 1))

            if self._K_horizontal > 0.0:
                magnetic_fields.append(SRWLMagFldH(1, 'h', self.magnetic_field_horizontal(), 0, -1, 1))

            return SRWLMagFldU(magnetic_fields,
                               self._period_length,
                               int(self._number_of_periods))

    def to_python_code_aux(self):
        if self._number_of_periods <=3:
            text_code = "n_points = 100" + "\n"

            text_code += "arBx = " + ("array('d', [0]*n_points)" if self._K_horizontal > 0.0 else "None") + "\n"
            text_code += "arBy = " + ("array('d', [0]*n_points)" if self._K_vertical > 0.0 else "None") + "\n"
            text_code += "period_length = " + str(self._period_length)
            text_code += "longitudinal_mesh = numpy.linspace(-period_length/2., period_length/2., n_points)" + "\n"
            text_code += "phases = 2.0*numpy.pi*(longitudinal_mesh/period_length)" + "\n"

            text_code += "for i in range (0, n_points):" + "\n"

            if self._K_horizontal > 0.0: text_code += "   arBx[i] = -" + str(self.magnetic_field_horizontal()) + " * numpy.cos(phases[i])" + "\n"
            if self._K_vertical > 0.0:   text_code += "   arBy[i] = " + str(self.magnetic_field_vertical()) + " * numpy.cos(phases[i])" + "\n"

            text_code += "magnetic_structure = SRWLMagFld3D(_arBx=arBx, _arBy=arBy, _nx=1, _ny=1, _nz=n_points, _rz=period_length, _nRep=" + str(int(self._number_of_periods)) + ", _interp=1)" + "\n"
        else:
            text_code = "magnetic_fields = []" + "\n"

            if self._K_vertical > 0.0:
                text_code += "magnetic_fields.append(SRWLMagFldH(1, 'v', " + str(self.magnetic_field_vertical()) + ", 0, 1, 1))" + "\n"

            if self._K_horizontal > 0.0:
                text_code += "magnetic_fields.append(SRWLMagFldH(1, 'h', " + str(self.magnetic_field_horizontal()) + ", 0, -1, 1))" + "\n"

            text_code += "magnetic_structure = SRWLMagFldU(magnetic_fields," + str(self._period_length) + "," + str(self._number_of_periods) + ")" + "\n"

        return text_code



import numpy

from syned.beamline.shape import Ellipse, Rectangle, Circle
from wofry.beamline.decorators import OpticalElementDecorator

from wofry.propagator.propagator import PropagationParameters
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters, WavefrontPropagationOptionalParameters
from wofrysrw.srw_object import SRWObject

from oasys_srw.srwlib import SRWLOptC, srwl, SRWLOptShift, SRWLOptAng

class Orientation:
    UP = 0
    DOWN = 1
    LEFT = 2
    RIGHT = 3


class SRWOpticalElementDisplacement(SRWObject):

    BEFORE=0
    AFTER=1

    def __init__(self, shift_x=0.0, shift_y=0.0, rotation_x=0.0, rotation_y=0.0):
        self.shift_x=shift_x
        self.shift_y=shift_y
        self.rotation_x=rotation_x
        self.rotation_y=rotation_y

        self.has_shift = self.shift_x != 0.0 or self.shift_y != 0.0
        self.has_rotation = self.rotation_x != 0.0 or self.rotation_y != 0.0

    def to_python_code(self, data=None):
        oe_name = data[0]
        where = data[2]

        if where==SRWOpticalElementDisplacement.BEFORE:
            sign = -1
            suffix = "before_"
            text_where = "Before"
        elif where==SRWOpticalElementDisplacement.AFTER:
            sign = 1
            suffix = "after_"
            text_where = "After"

        text_code = ""

        if self.has_shift:
            text_code += "# " + oe_name + " Displacement (" + text_where + ")\n\n"

            text_code += "shift_" + suffix + oe_name + " = SRWLOptShift(_shift_x=" + str(sign*self.shift_x) + ", _shift_y=" + str(sign*self.shift_y) + ")" + "\n"
            text_code += "pp_shift_" + suffix + oe_name + " = " + WavefrontPropagationParameters().to_python_code() + "\n"

        if self.has_rotation:
            if self.has_shift: text_code += "\n"
            else: text_code += "# " + oe_name + " Displacement (" + text_where + ")\n\n"
            text_code += "rotation_" + suffix + oe_name + " = SRWLOptAng(_ang_x=" + str(sign*self.rotation_x) + ", _ang_y=" + str(sign*self.rotation_y) + ")" + "\n"
            text_code += "pp_rotation_" + suffix + oe_name + " = " + WavefrontPropagationParameters().to_python_code() + "\n"

        return text_code

class SRWOpticalElementDecorator(SRWObject):
    def toSRWLOpt(self):
        raise NotImplementedError("")

    def fromSRWLOpt(self, srwlopt=None):
        raise NotImplementedError("")

class SRWOpticalElement(SRWOpticalElementDecorator, OpticalElementDecorator):

    def __init__(self, optical_element_displacement=None):
        self.displacement = optical_element_displacement

    def add_displacement_to_array(self, oe_array, pp_array, where=SRWOpticalElementDisplacement.BEFORE):

        sign = -1 if where==SRWOpticalElementDisplacement.BEFORE else 1

        if self.displacement.has_shift:
            oe_array.append(SRWLOptShift(_shift_x=sign*self.displacement.shift_x, _shift_y=sign*self.displacement.shift_y))
            pp_array.append(self.get_default_propagation_parameters())

        if self.displacement.has_rotation:
            oe_array.append(SRWLOptAng(_ang_x=sign*self.displacement.rotation_x, _ang_y=sign*self.displacement.rotation_y))
            pp_array.append(self.get_default_propagation_parameters())

    def applyOpticalElement(self, wavefront=None, parameters=None, element_index=None):
        oe_array = []
        pp_array = []

        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.BEFORE)

        oe_array.append(self.toSRWLOpt())
        pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters))

        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.AFTER)

        optBL = SRWLOptC(oe_array,
                         pp_array)

        srwl.PropagElecField(wavefront, optBL)

        return wavefront

    def add_to_srw_native_array(self, oe_array = [], pp_array=[], parameters=None, wavefront=None):
        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.BEFORE)

        oe_array.append(self.toSRWLOpt())
        pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters))

        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.AFTER)

    def get_srw_wavefront_propagation_parameter(self, parameters):
        if isinstance(parameters, PropagationParameters):
            if not parameters.has_additional_parameter("srw_oe_wavefront_propagation_parameters"):
                wavefront_propagation_parameters = WavefrontPropagationParameters()
            else:
                wavefront_propagation_parameters = parameters.get_additional_parameter("srw_oe_wavefront_propagation_parameters")

                if not isinstance(wavefront_propagation_parameters, WavefrontPropagationParameters):
                    raise ValueError("SRW Wavefront Propagation Parameters are inconsistent")

            srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()

            if parameters.has_additional_parameter("srw_oe_wavefront_propagation_optional_parameters"):
                wavefront_propagation_optional_parameters = parameters.get_additional_parameter("srw_oe_wavefront_propagation_optional_parameters")

                if not isinstance(wavefront_propagation_optional_parameters, WavefrontPropagationOptionalParameters):
                    raise ValueError("SRW Wavefront Propagation Optional Parameters are inconsistent")

                wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)
        elif isinstance(parameters, list):
            wavefront_propagation_parameters = parameters[0]
            wavefront_propagation_optional_parameters = parameters[1]

            srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()
            if not wavefront_propagation_optional_parameters is None: wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)

        return srw_parameters_array

    def getXY(self):
        boundary_shape = self.get_boundary_shape()

        if not boundary_shape is None:
            if isinstance(boundary_shape, Rectangle) or isinstance(boundary_shape, Ellipse):
                x_left, x_right, y_bottom, y_top = boundary_shape.get_boundaries()

                return x_left + 0.5*(x_right-x_left), \
                       y_bottom + 0.5*(y_top-y_bottom)

            elif isinstance(boundary_shape, Circle):
                radius, x_center, y_center = boundary_shape.get_boundaries()

                return x_center, y_center
        else:
            return 0.0, 0.0

    def get_orientation_vectors(self):
        sign = (-1 if self.invert_tangent_component else 1)

        if self.orientation_of_reflection_plane == Orientation.LEFT:
            nvx = -numpy.cos(self.grazing_angle)
            nvy = 0
            nvz = -numpy.sin(self.grazing_angle)
            tvx = sign*numpy.sin(self.grazing_angle)
            tvy = 0
        elif self.orientation_of_reflection_plane == Orientation.RIGHT:
            nvx = numpy.cos(self.grazing_angle)
            nvy = 0
            nvz = -numpy.sin(self.grazing_angle)
            tvx = -sign*numpy.sin(self.grazing_angle)
            tvy = 0
        elif self.orientation_of_reflection_plane == Orientation.UP:
            nvx = 0
            nvy = numpy.cos(self.grazing_angle)
            nvz = -numpy.sin(self.grazing_angle)
            tvx = 0
            tvy = -sign*numpy.sin(self.grazing_angle)
        elif self.orientation_of_reflection_plane == Orientation.DOWN:
            nvx = 0
            nvy = -numpy.cos(self.grazing_angle)
            nvz = -numpy.sin(self.grazing_angle)
            tvx = 0
            tvy = -sign*numpy.sin(self.grazing_angle)

        return nvx, nvy, nvz, tvx, tvy

    def get_default_propagation_parameters(self):
        return WavefrontPropagationParameters().to_SRW_array()

from oasys_srw.srwlib import SRWLOptA

class SRWOpticalElementWithAcceptanceSlit(SRWOpticalElement):

    def __init__(self,
                 optical_element_displacement         = None,
                 tangential_size                      = 1.2,
                 sagittal_size                        = 0.01,
                 vertical_position_of_mirror_center   = 0.0,
                 horizontal_position_of_mirror_center = 0.0,
                 grazing_angle                        = 0.003,
                 orientation_of_reflection_plane      = Orientation.UP,
                 invert_tangent_component             = False,
                 add_acceptance_slit                  = False):

        super(SRWOpticalElementWithAcceptanceSlit, self).__init__(optical_element_displacement)

        self.vertical_position_of_mirror_center = vertical_position_of_mirror_center
        self.horizontal_position_of_mirror_center = horizontal_position_of_mirror_center

        self.tangential_size                                  = tangential_size
        self.sagittal_size                                    = sagittal_size
        self.grazing_angle                                    = grazing_angle
        self.orientation_of_reflection_plane                  = orientation_of_reflection_plane
        self.invert_tangent_component                         = invert_tangent_component

        self.add_acceptance_slit=add_acceptance_slit

    def get_acceptance_slit(self):
        if self.orientation_of_reflection_plane == Orientation.UP or \
                self.orientation_of_reflection_plane==Orientation.DOWN:
            vertical_aperture   = self.tangential_size*numpy.sin(self.grazing_angle)
            horizontal_aperture = self.sagittal_size
        else:
            vertical_aperture   = self.sagittal_size
            horizontal_aperture = self.tangential_size*numpy.sin(self.grazing_angle)

        return SRWLOptA('r', 'a', horizontal_aperture, vertical_aperture)

    def create_propagation_elements(self):
        optical_elements = []
        propagation_parameters = []

        if not self.displacement is None and self.displacement.where == SRWOpticalElementDisplacement.BEFORE:
            self.add_displacement_to_array(optical_elements, propagation_parameters)

        if not self.add_acceptance_slit:
            optical_elements.append(self.toSRWLOpt())
            propagation_parameters.append(self.get_srw_wavefront_propagation_parameter())
        else:
            optical_elements.extend([self.get_acceptance_slit(), self.toSRWLOpt()])
            propagation_parameters.extend([self.get_srw_wavefront_propagation_parameter(), # all the resizing/resampling goes to the slit
                                           self.get_default_propagation_parameters()]) # no resizing/resampling needed

        if not self.displacement is None and self.displacement.where == SRWOpticalElementDisplacement.AFTER:
            self.add_displacement_to_array(optical_elements, propagation_parameters)

        return optical_elements, propagation_parameters

    def add_to_srw_native_array(self, oe_array = [], pp_array=[], parameters=None, wavefront=None):
        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.BEFORE)

        if not self.add_acceptance_slit:
            oe_array.append(self.toSRWLOpt())
            pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters))
        else:
            oe_array.append(self.get_acceptance_slit())
            oe_array.append(self.toSRWLOpt())
            pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters)) # all the resizing/resampling goes to the slit
            pp_array.append(self.get_default_propagation_parameters()) # no resizing/resampling needed

        if not self.displacement is None:
            self.add_displacement_to_array(oe_array, pp_array, where=SRWOpticalElementDisplacement.AFTER)



from syned.beamline.shape import Circle

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement

from oasys_srw.srwlib import *

class SRWZonePlate(SRWOpticalElement):
    def __init__(self,
                 name="Undefined",
                 optical_element_displacement=None,
                 total_number_of_zones=100,
                 outer_zone_radius=0.1e-03,
                 thickness=10e-06,
                 delta_main_material=1e-06,
                 delta_complementary_material=0.0,
                 attenuation_length_main_material=0.1,
                 attenuation_length_complementary_material=1e-06,
                 x=0.0,
                 y=0.0):
        SRWOpticalElement.__init__(self, optical_element_displacement=optical_element_displacement)

        self._name=name

        self.total_number_of_zones=total_number_of_zones
        self.outer_zone_radius=outer_zone_radius
        self.thickness=thickness
        self.delta_main_material=delta_main_material
        self.delta_complementary_material=delta_complementary_material
        self.attenuation_length_main_material=attenuation_length_main_material
        self.attenuation_length_complementary_material=attenuation_length_complementary_material
        self.x = x
        self.y = y

    def toSRWLOpt(self):
        return SRWLOptZP(_nZones=self.total_number_of_zones,
                         _rn=self.outer_zone_radius,
                         _thick=self.thickness,
                         _delta1=self.delta_main_material,
                         _atLen1=self.attenuation_length_main_material,
                         _delta2=self.delta_complementary_material,
                         _atLen2=self.attenuation_length_complementary_material,
                         _x=self.x,
                         _y=self.y)

    def fromSRWLOpt(self, srwlopt=SRWLOptZP()):
        self.total_number_of_zones = srwlopt.nZones
        self.outer_zone_radius = srwlopt.rn
        self.thickness=srwlopt.thick
        self.delta_main_material=srwlopt.delta1
        self.delta_complementary_material=srwlopt.delta2
        self.attenuation_length_main_material=srwlopt.atLen1
        self.attenuation_length_complementary_material=srwlopt.atLen2
        self.x = srwlopt.x
        self.y = srwlopt.y

    def to_python_code(self, data=None):
        text_code = data[0] + "="+  "SRWLOptZP(_nZones=" + str(self.total_number_of_zones) + "," +  "\n"
        text_code += "                _rn=" + str(self.outer_zone_radius) + "," +  "\n"
        text_code += "                _thick=" + str(self.thickness) + "," +  "\n"
        text_code += "                _delta1=" + str(self.delta_main_material) + "," +  "\n"
        text_code += "                _atLen1=" + str(self.attenuation_length_main_material) + "," +  "\n"
        text_code += "                _delta2=" + str(self.delta_complementary_material) + "," +  "\n"
        text_code += "                _atLen2=" + str(self.attenuation_length_complementary_material) + "," +  "\n"
        text_code += "                _x=" + str(self.x) + "," +  "\n"
        text_code += "                _y=" + str(self.y) + ")" + "\n"

        return text_code

    def get_boundary_shape(self):
        return Circle(a_axis_min=self.outer_zone_radius,
                      x_center=self.x,
                      y_center=self.y)
